/***
 * This is the main file for the sysdiag utility.
 *
 *
 * Copyright (C) 2003 International Business Machines, Inc.
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation version 2 of the License.
 *
 *	This program is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *	General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License along
 *	with this program; if not, write to the Free Software Foundation, Inc.,
 *	675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * 	need api in libsysfs to get one attribute of one particular device.
 *	also an api to get all attributes of one device.
 *
 */

#include "sysdiag.h"


/**
 * displays long usage.
 */
int usage()
{

	fprintf(stdout, "sysdiag [<options> [dev-name]]\n\n");
	fprintf(stdout, " options:\n");
	fprintf(stdout, "\t-h\t\t\tusage\n");
	fprintf(stdout, "\t-d\t\t\tShow diagnostics capabilites\n\n");
	fprintf(stdout, "\t-i\t\t\trun device inquiry cmd\n");
	fprintf(stdout, "\t-s\t\t\tshow device statistics\n");
	fprintf(stdout, "\t-t\t\t\trun diagnostic test cmd\n");
	fprintf(stdout, "\t-l\t\t\tshow logs\n\n");
	fprintf(stdout, "\t-x\t\t\trun diag on all devices\n");
	fprintf(stdout, "\t-p\t\t\tstop device before running diag\n");
	fprintf(stdout, "\t-b <bus-name>\t\tspecify bus to use\n");
	fprintf(stdout, "\t-c <class-name>\t\tspecify class to use\n");
	fprintf(stdout, "\t-r <root-dev>\t\tspecify root device to show\n");
	fprintf(stdout, "\t-v\t\t\tmore verbose\n");
	fprintf(stdout, "\ndev-name:\teth0, sg0, XXXX:XX:XX.X (pci busid), X:X:X:X (scsi id) etc.\n");
	fprintf(stdout, "bus-name:\tpci, scsi, etc.\n");
	fprintf(stdout, "class-name:\tnet, scsi_host, etc.\n");
	fprintf(stdout, "root-dev:\tpci0000:00, legacy, etc.\n");
	return 0;
}

void show_diag_info(char *device)
{
	struct diags *diag = NULL, *curr = NULL;

	diag = parse_main(diag);
	curr = diag;
	
	while (curr != NULL)
	{
		if (device != NULL) {
			if (strncasecmp(curr->name, device, 
						strlen(device)) != 0) {
				curr = curr->next;
				continue;
			}
		}
		fprintf(stdout, "\n-------------------------------------\n");
		fprintf(stdout, "Diagnostics for class %s:\n", curr->name);
		fprintf(stdout, "-------------------------------------\n");
		if (curr->test != NULL) {
			fprintf(stdout, "Test diag option      :  -t"
					"\t%s\n", curr->test);
		}
		if (curr->inq != NULL) {
			fprintf(stdout, "Inquiry Diag option   :  -i"
					"\t%s\n", curr->inq);
		}
		if (curr->stat != NULL) {
			fprintf(stdout, "Statistic Diag option :  -s"
					"\t%s\n", curr->stat);
		}
		curr = curr->next;
	}
	if (diag != NULL)
		free_diag(diag);
	
}


/**
 * This will setup the various sysfs paths.
 */
int setup_sysfs_paths()
{
	memset(SYSFS_MOUNT_PATH, 0, SYSFS_PATH_MAX);
	if ( sysfs_get_mnt_path(SYSFS_MOUNT_PATH, SYSFS_PATH_MAX) != 0) {
		fprintf(stderr,"sysfs not supported on your system\n");
		exit(1);
	}
	return 0;

}




/**
 * get_scsi_quad().
 * given a device name of sdX, or sg0 or a quad, it return the
 * /dev/sgX for the sg3utils to use.
 *
 * if the device name passed is null then we return the first device
 * we get from the sg_map.
 *
 */
char * get_scsi_quad(char *device)
{
	int fd, size, i;
	char *buf, *ptr1, *ptr2, *ptr3, *temp, *dev_num;
	char quad[] = "0:0:0:0";
	struct stat stat_buf;
	
	if ((DEVICE_NAME[0] != 0) && (strlen(device) < 3))
		return NULL;

	if (geteuid() != 0) {
		fprintf(stderr, "need to log in as root!\n");
		exit(0);
	}
	system(SG_MAP);
	if (stat(SYSDIAG_FILE, &stat_buf) != 0) {
		fprintf(stderr, "error reading file %s\n",SYSDIAG_FILE);
		return NULL;
	}
	size = stat_buf.st_size;
	buf = (char *)calloc(size, sizeof(char));
	ptr3 = ptr1 = buf;

	if ((fd = open(SYSDIAG_FILE, O_RDONLY)) == -1) {
		free(buf);
		return NULL;
	}

	if ((size = read(fd, buf, size)) == -1) {
		free(buf);
		close(fd);
		return NULL;
	}
	close(fd);
	ptr2 = NULL;
	dev_num = NULL;
	while ((ptr2 = strchr(ptr1, '\n'))!= NULL) {
		size = ptr2 - ptr1;
		temp = (char *)calloc(size, sizeof(char));
		strncpy(temp, ptr1, size);
		ptr1 = ptr2 + 1;
		ptr2 = strstr(temp, " ");
		size = ptr2 - temp;
		dev_num = (char *)calloc(size+1, sizeof(char));
		strncpy(dev_num, temp, size);
		if (device == NULL) {
			free(temp);
			free(buf);
			return dev_num;
		}
		ptr3 = strstr(ptr2, " ");
		ptr3+=2;
		i = 0;
		while ((ptr3 != NULL)) {
			quad[i] = ptr3[0];
			ptr3 = strstr(ptr3, " ");
			ptr3++;
			i+=2;
			if (i > 6)
				break;
		}
		if (device[1] == ':') {
			if ((strncmp(quad, device, 7)) == 0) {
				free(temp);
				free(buf);
				return dev_num;
			}
		}
		else {
			if (strstr(temp,device) != NULL) {
				strncpy(DEVICE_NAME, quad, strlen(quad));
				strncpy(BUS_NAME, "scsi", strlen("scsi"));
				DIAG_OPTIONS |= DIAG_BUS | DIAG_DEVICE;
				free(temp);
				free(buf);
				return dev_num;
			}
		}
		ptr2 = NULL;
	}
	free(buf);
	free(temp);
	return NULL;
}

char *get_def_scsi_host_dev(char *device)
{
	char *path = NULL, *dev_num = NULL;
	struct sysfs_directory *dir = NULL, *current = NULL;
	struct stat statbuf;
	if (device[0] == 0)
		return NULL;
	
	path = (char *)calloc(SYSFS_PATH_MAX, sizeof(char));
	
	strncpy(path, SYSFS_MOUNT_PATH, strlen(SYSFS_MOUNT_PATH));
	strncat(path, SYSFS_CLASS_DIR, strlen(SYSFS_CLASS_DIR));
	strncat(path, "/scsi_host/", 11);
	strncat(path, DEVICE_NAME, strlen(DEVICE_NAME));
	strncat(path, "/device", 7);
	if (stat(path, &statbuf) != 0) { 
		free (path);
		return 0;
	}
	dir = sysfs_open_directory(path);
	if (dir == NULL) {
		fprintf(stderr, "Error opening sysfs directory at %s\n", path);
		return NULL;
	}
	if ((sysfs_read_directory(dir)) != 0) {
		fprintf(stderr, "Error reading sysfs directory at %s\n", path);
		return NULL;
	}
	current = dir->subdirs;
	memset(path, 0, SYSFS_PATH_MAX);
	while (current != NULL) {
		if ((sysfs_get_name_from_path(current->path, path,
						SYSFS_NAME_LEN)) != 0) {
			sysfs_close_directory(dir);
			return NULL;
		}
		dev_num = get_scsi_quad(path);
		break;
	}
	sysfs_close_directory(dir);
	return dev_num;
}


char *make_diag_path(char *path, char *dev_num)
{
	char *ptr = NULL;
	char *diag_path = NULL;
	int size = 0;

	size = strlen(dev_num);
	if ((ptr = strstr(path, "<dev>")) != NULL) {
		size += strlen(path);
		size-=5;
		diag_path = (char *)calloc(size+1, sizeof(char));
		strncpy(diag_path, path, (ptr - path));
		strncat(diag_path, dev_num, strlen(dev_num));
		ptr+=5;
		strncat(diag_path, ptr, strlen(ptr));
	}
	else {
		size += strlen(path);
		diag_path = (char *)calloc(size+1, sizeof(char));
		strncpy(diag_path, path, strlen(path));
		strncat(diag_path, " ", 1);
		strncat(diag_path, dev_num, strlen(dev_num));
	}
	return (char *)diag_path;

}

/**
 * This function based on the option will call the relavent
 * sg3utils applications.
 * dev_num is in sgX format.
 */
int run_scsi_diag(char *dev_num)
{
	char *path = NULL;
	char num[10];
	int size = 0;
	struct diags *diag_head = NULL, *curr_diag = NULL;

	diag_head = parse_main(diag_head);
	curr_diag = diag_head;
	while (curr_diag!= NULL) {
		if (strcmp(curr_diag->name, "scsi_host") == 0)
			break;
		curr_diag=curr_diag->next;
	}
	size = strlen(dev_num);
	if (DIAG_OPTIONS & DIAG_INQ) {
		if (curr_diag == NULL || curr_diag->inq == NULL) {
			size += strlen(SCSI_INFO);
			path = (char *)calloc(size, sizeof(char));
			strncpy(path, SCSI_INFO, strlen(SCSI_INFO));
			strncat(path, dev_num, strlen(dev_num));
		} else 	if (curr_diag!= NULL) {
			if (curr_diag->inq != NULL)
				path = make_diag_path(curr_diag->inq, dev_num);
		}
		fprintf(stdout, "\n-------------------------------------\n");
		fprintf(stdout, "Device inquiry on : %s\n", dev_num);
		fprintf(stdout, "-------------------------------------\n");
		fflush(stdout);
		system(path);
	}
	if (DIAG_OPTIONS & DIAG_STAT) {
		if (curr_diag == NULL || curr_diag->stat == NULL) {
			size += strlen(SCSI_STAT);
			path = (char *)calloc(size, sizeof(char));
			strncpy(path, SCSI_STAT, strlen(SCSI_STAT));
			strncat(path, dev_num, strlen(dev_num));
		} else 	if (curr_diag!= NULL) {
			if (curr_diag->stat != NULL)
				path = make_diag_path(curr_diag->stat, dev_num);
		}
		fprintf(stdout, "\n-------------------------------------\n");
		fprintf(stdout, "Device statistics of : %s\n", dev_num);
		fprintf(stdout, "-------------------------------------\n");
		fflush(stdout);
		system(path);
	}
	if (DIAG_OPTIONS & DIAG_TEST) {
		fprintf(stdout, "\n-------------------------------------\n");
		fprintf(stdout, "Testing Device : %s\n", dev_num);
		fprintf(stdout, "-------------------------------------\n");
		if (curr_diag == NULL || curr_diag->test == NULL) {
			size += strlen(SCSI_TEST);
			path = (char *)calloc(size, sizeof(char));
			strncpy(path, SCSI_TEST, strlen(SCSI_TEST));
			strncat(path, dev_num, strlen(dev_num));
			memset(num,0,10);
			fprintf(stdout, "Testing SCSI RW Buf, enter the byte size to test for:");
			fflush(stdin);
			scanf("%s",num);
			fprintf(stderr, "the size is::%s\n", num);
			strncat(path, " ", 1);
			strncat(path, num, strlen(num));
		} else 	if (curr_diag!= NULL) {
			if (curr_diag->test != NULL)
				path = make_diag_path(curr_diag->test, dev_num);
		}
		if (DIAG_OPTIONS & DIAG_PAUSE) {
			char *down = NULL;
			size = strlen(dev_num) + 15;
			down = (char *)calloc(size, sizeof(char));
			strncpy(down, "sg_start  ", 10);
			strncat(down, dev_num, strlen(dev_num));
			strncat(down, "  0", 3);
			fprintf(stdout, "Bringing down device:%s\n", dev_num);
			fflush(stdout);
			system(down);
			free(down);
		}
		fflush(stdout);
		system(path);
		if (DIAG_OPTIONS & DIAG_PAUSE) {
			char *up = NULL;
			size = strlen(dev_num) + 15;
			up = (char *)calloc(size, sizeof(char));
			strncpy(up, "sg_start  ", 10);
			strncat(up, dev_num, strlen(dev_num));
			strncat(up, "  1", 3);
			fprintf(stdout, "Bringing up device:%s\n", dev_num);
			fflush(stdout);
			system(up);
			free(up);
		}
	}
	if (DIAG_OPTIONS & DIAG_LOGS) {
		path = (char *)show_logs("scsi");
		fprintf(stdout, "\n-------------------------------------\n");
		fprintf(stdout, "Error logs for Device : %s\n", dev_num);
		fprintf(stdout, "-------------------------------------\n");
		fflush(stdout);
		system(path);
	}
	if (path != NULL)
		free(path);
	if (diag_head != NULL)
		free_diag(diag_head);
	return 0;
}


/**
 * run_scsi_diag_all().
 * this runs diag on all the scsi devices.
 * if DEVICE_NAME is not null
 *	then we have to run diag on all devices on the that
 * 	host bus adapter, i.e. host0, host1 etc. For this we
 *	go to /sys/class/scsi_host/host0/device directory and
 *	and run diag on all the devices there.
 *
 * else
 *	we go to the sys/bus/scsi/devices/ dir and run diag on
 *	all devices there.
 *
 *
 *
 */
int run_scsi_diag_all()
{
	char *path = NULL;
	struct sysfs_directory *dir = NULL, *current = NULL;
	char *dev_num = NULL;
	struct stat statbuf;

	path = (char *)calloc(SYSFS_PATH_MAX, sizeof(char));

	if (DEVICE_NAME[0] == 0) {
		strncpy(path, SYSFS_MOUNT_PATH, strlen(SYSFS_MOUNT_PATH));
		strncat(path, SYSFS_CLASS_DIR, strlen(SYSFS_CLASS_DIR));
		strncat(path, "/scsi_device", 12);
	}
	else {
		strncpy(path, SYSFS_MOUNT_PATH, strlen(SYSFS_MOUNT_PATH));
		strncat(path, SYSFS_CLASS_DIR, strlen(SYSFS_CLASS_DIR));
		strncat(path, "/scsi_host/", 11);
		strncat(path, DEVICE_NAME, strlen(DEVICE_NAME));
		strncat(path, "/device", 7);
	}
	if (stat(path, &statbuf) != 0) { 
		free (path);
		return 0;
	}
	dir = sysfs_open_directory(path);
	if (dir == NULL) {
		fprintf(stderr, "Error opening sysfs directory at %s\n", path);
		return 1;
	}
	if ((sysfs_read_directory(dir)) != 0) {
		fprintf(stderr, "Error reading sysfs directory at %s\n", path);
		return 1;
	}
	current = dir->subdirs;
	memset(path, 0, SYSFS_PATH_MAX);
	while (current != NULL) {
		if ((sysfs_get_name_from_path(current->path, path,
						SYSFS_NAME_LEN)) != 0) {
			sysfs_close_directory(dir);
			return 1;
		}
		dev_num = get_scsi_quad(path);
		run_scsi_diag(dev_num);
		current = current->next;
	}
	free(path);
	free(dev_num);
	sysfs_close_directory(dir);
	return 0;
}


/**
 * this will call the ethtool with the relevant options.
 * the dev_num is in ethx format.
 */
int run_net_diag(char *dev_num)
{
	char *path = NULL;
	int size = 0;
	struct diags *diag_head = NULL, *curr_diag = NULL;

	diag_head = parse_main(diag_head);
	curr_diag = diag_head;
	while (curr_diag!= NULL) {
		if (strcmp(curr_diag->name, "net") == 0)
			break;
		curr_diag=curr_diag->next;
	}
	size  = strlen(dev_num);

	if (DIAG_OPTIONS & DIAG_INQ) {
		if ((curr_diag == NULL) || (curr_diag->inq == NULL)) {
			size  += strlen(NET_INFO);
			path = (char *)calloc(size, sizeof(char));
			strncpy(path, NET_INFO, strlen(NET_INFO));
			strncat(path, dev_num, strlen(dev_num));
		} else 	if ( curr_diag!= NULL) {
			if (curr_diag->inq != NULL) 
				path = make_diag_path(curr_diag->inq, dev_num);
			
		}
		fprintf(stdout, "\n-------------------------------------\n");
		fprintf(stdout, "Device inquiry on : %s\n", dev_num);
		fprintf(stdout, "-------------------------------------\n");
		fflush(stdout);
		system(path);
	}
	if (DIAG_OPTIONS & DIAG_STAT) {
		if ((curr_diag == NULL) || (curr_diag->test == NULL)) {
			size  += strlen(NET_STAT);
			path = (char *)calloc(size, sizeof(char));
			strncpy(path, NET_STAT, strlen(NET_STAT));
			strncat(path, dev_num, strlen(dev_num));
		} else	if (curr_diag!= NULL) {
			if (curr_diag->stat != NULL)
				path = make_diag_path(curr_diag->stat, dev_num);
		}
		fprintf(stdout, "\n-------------------------------------\n");
		fprintf(stdout, "Device statistics of : %s\n", dev_num);
		fprintf(stdout, "-------------------------------------\n");
		fflush(stdout);
		system(path);
	}
	if (DIAG_OPTIONS & DIAG_TEST) {
		if ((curr_diag == NULL) || (curr_diag->test == NULL)) {
			size  += strlen(NET_TEST);
			path = (char *)calloc(size, sizeof(char));
			strncpy(path, NET_TEST, strlen(NET_TEST));
			strncat(path, dev_num, strlen(dev_num));
		} else 	if (curr_diag!= NULL) {
			if (curr_diag->test != NULL)
				path = make_diag_path(curr_diag->test, dev_num);
		}
		fprintf(stdout, "\n-------------------------------------\n");
		fprintf(stdout, "Testing Device : %s\n", dev_num);
		fprintf(stdout, "-------------------------------------\n");
		if (DIAG_OPTIONS & DIAG_PAUSE) {
			char *down = NULL;
			size = strlen(dev_num) + 15;
			down = (char *)calloc(size, sizeof(char));
			strncpy(down, "ifconfig ", 9);
			strncat(down, dev_num, strlen(dev_num));
			strncat(down, " down", 5);
			fprintf(stdout, "Bringing down device:%s\n", dev_num);
			fflush(stdout);
			system(down);
			free(down);
		}
		fflush(stdout);
		system(path);

		if (DIAG_OPTIONS & DIAG_PAUSE) {
			char *up = NULL;
			size = strlen(dev_num) + 13;
			up = (char *)calloc(size, sizeof(char));
			strncpy(up, "ifconfig ", 9);
			strncat(up, dev_num, strlen(dev_num));
			strncat(up, " up",3);
			fprintf(stdout, "Bringing up device:%s\n", dev_num);
			fflush(stdout);
			system(up);
			free(up);
		}
	}
	if (DIAG_OPTIONS & DIAG_LOGS) {
		path = (char *)show_logs((char *)dev_num);
		fprintf(stdout, "\n-------------------------------------\n");
		fprintf(stdout, "Error logs for Device : %s\n", dev_num);
		fprintf(stdout, "-------------------------------------\n");
		fflush(stdout);
		system(path);
	}
	if (path == NULL) {
		exit(1);
	}
	if (diag_head != NULL)
		free_diag(diag_head);
	free(path);
	return 0;
}


/**
 * run_net_diag_all():
 * This will work with only the class option,
 * goto the sys/class/net dir and run diag on all devices
 * there, check if the name of the device is in the ethX format,
 * to take care of the lo and dummy driver case.
 *
 *
 */
int run_net_diag_all(void)
{
	char *path = NULL, *name = NULL;
	struct sysfs_directory *dir = NULL, *current = NULL;
	struct stat statbuf;

	path = (char *)calloc(SYSFS_PATH_MAX, sizeof(char));
	strncpy(path, SYSFS_MOUNT_PATH, strlen(SYSFS_MOUNT_PATH));
	strncat(path, SYSFS_CLASS_DIR, strlen(SYSFS_CLASS_DIR));
	strncat(path, "/net", 4);

	if (stat(path, &statbuf) != 0) { 
		free (path);
		return 0;
	}

	dir = sysfs_open_directory(path);
	if (dir == NULL) {
		fprintf(stderr, "Error opening sysfs directory at %s\n", path);
		return 1;
	}
	if ((sysfs_read_directory(dir)) != 0) {
		fprintf(stderr, "Error reading sysfs directory at %s\n", path);
		return 1;
	}
	name = (char *)calloc(SYSFS_PATH_MAX, sizeof(char));
	for (current = dir->subdirs; current!=NULL; current=current->next)
	{
		if ((sysfs_get_name_from_path(current->path, name,
						SYSFS_NAME_LEN)) != 0) {
			sysfs_close_directory(dir);
			return 1;
		}
		if (strncmp(name, "eth", 3) != 0) {
			continue;
		}
		run_net_diag(name);
	}
	sysfs_close_directory(dir);
	free(path);
	free(name);
	return 0;
}


/**
 * given device-name it will check if it is a valid device name and
 * also get the valid device name in case the format is diff than
 * ethX.
 * in case of net class the device name will be in ethX format.
 * we just need to validate it.
 *
 * the network diags are only with the class option.
 * option -c net -D ethX or 0000:00:00.0, or 00:00.0
 * but doesnot work with the pci option, well we can make it with
 * the pci option, since right now only the device we will run diag
 * thro the bus option will be a network device not scsi device.
 *
 * for pci only network devices will be test. If a scsi device
 * bus_id is passed in -b pci option it will not work.
 * Need to pass that with -b scsi option.
 *
 * in case of pci, the format will be 0000:00:00.0 and also it can
 * so we need to check if that device is a network device.
 * check of class of device is 02.
 * in sysfs format for class is 020000, firt byte is the class.
 * so take first two chars and cmp with 02.
 *
 */
char *get_net_dev_num(char * dev_name)
{
	char *path = NULL, *name = NULL;
	struct sysfs_directory *dir = NULL, *current = NULL;

	path = (char *)calloc(SYSFS_PATH_MAX, sizeof(char));
	strncpy(path, SYSFS_MOUNT_PATH, strlen(SYSFS_MOUNT_PATH));
	strncat(path, SYSFS_CLASS_DIR, strlen(SYSFS_CLASS_DIR));
	strncat(path, "/net", 4);

	if ((dir = sysfs_open_directory(path)) == NULL) {
		fprintf(stderr,"open dir failed\n");
		free(path);
		return NULL;
	}
	free(path);
	if ((sysfs_read_directory(dir)) != 0) {
		fprintf( stderr, "Error reading sysfs directory at %s\n",path);
		sysfs_close_directory(dir);
		return NULL;
	}
	name = (char *)calloc(SYSFS_PATH_MAX, sizeof(char));
	for (current=dir->subdirs; current != NULL; current=current->next) {

		if ((sysfs_get_name_from_path(	current->path,
						name,
						SYSFS_NAME_LEN))
						!= 0) {
			sysfs_close_directory(dir);
			return NULL;
		}
		if (strncmp(name, "eth",3) != 0)
			continue;
		if (dev_name == NULL && DEVICE_NAME[0] == 0) {
			sysfs_close_directory(dir);
			return name;
		}
		if (strncmp(name, DEVICE_NAME, strlen(DEVICE_NAME)) == 0) {
			sysfs_close_directory(dir);
			return name;
		}
	}
	sysfs_close_directory(dir);
	return NULL;
}


char *get_net_bus_dev(char *dev_name)
{

	struct sysfs_directory *dir = NULL, *current = NULL;
	struct sysfs_directory *subdir = NULL;
	char *path = NULL, *name = NULL;

	dev_name = (char *)calloc(SYSFS_PATH_MAX, sizeof(char));

	if (DEVICE_NAME[2] == ':')
		strncpy(dev_name,"0000:",5);
	strncat(dev_name, DEVICE_NAME, strlen(DEVICE_NAME));

	path = (char *)calloc(SYSFS_PATH_MAX, sizeof(char));
	strncpy(path, SYSFS_MOUNT_PATH, strlen(SYSFS_MOUNT_PATH));
	strncat(path, SYSFS_CLASS_DIR, strlen(SYSFS_CLASS_DIR));
	strncat(path, "/net",4);
	if ((dir = sysfs_open_directory(path)) == NULL ) {
		fprintf(stderr,"sysfs_open_dir failed for %s\n",path);
		return NULL;
	}
	if (sysfs_read_directory(dir) != 0) {
		sysfs_close_directory(dir);
		fprintf(stderr,"sysfs_read_dir failed for %s\n",path);
		return NULL;
	}
	name = malloc(SYSFS_PATH_MAX);
	for (current = dir->subdirs; current != NULL; current = current->next)
	{
		memset(name, 0, SYSFS_PATH_MAX);
		if ((sysfs_get_name_from_path( 	current->path, name,
						SYSFS_NAME_LEN)) != 0) {
			sysfs_close_directory(dir);
			return NULL;
		}
		if (strncmp(name, "eth", 3) != 0)
			continue;
		if ((subdir = sysfs_open_directory(current->path)) == NULL) {
			fprintf(stderr, "sysfs_open_dir failed for %s\n",
					current->path);
			return NULL;
		}
		if (sysfs_read_directory(subdir) != 0) {
			fprintf(stderr, "sysfs_read_dir failed for %s\n",
					current->path);
			return NULL;
		}
		if (subdir->links != NULL) {
			struct sysfs_link *curl = subdir->links;
			while (curl != NULL) {
				if (strstr(curl->target, dev_name) != NULL)
					return (char *)name;
				curl = curl->next;
			}
		}
	}
	sysfs_close_directory(dir);
	return NULL;
}



char *get_net_dev(char *dev_num)
{

	if (DEVICE_NAME != NULL) {
		if (DEVICE_NAME[2] == ':' || DEVICE_NAME[4] == ':') {
			if ((dev_num = get_net_bus_dev(dev_num)) == NULL) {
				return NULL;
			}
			return (char *)dev_num;
		}
	}
	if ((dev_num = get_net_dev_num(dev_num)) == NULL) {
		fprintf(stderr, "failed to get network device name\n");
		return NULL;
	}
	return (char *)dev_num;

}


int run_diag_bus()
{
	char *dev_num = NULL;
	char *path = NULL;

	if (strncmp(BUS_NAME, "scsi", 4) == 0) {
		if (DIAG_OPTIONS & DIAG_ALL) {
			if (DIAG_OPTIONS & DIAG_LOGS) {
		fprintf(stdout, "\n-------------------------------------\n");
		fprintf(stdout, "Getting error logs for SCSI\n");
		fprintf(stdout, "-------------------------------------\n");
				DIAG_OPTIONS &= ~DIAG_LOGS;
				path = (char *)show_logs("scsi");
				if (path != NULL) {
					fflush(stdout);
					system(path);
				}
			}
			return (run_scsi_diag_all());
		}
		if ((dev_num = get_scsi_quad((char *)DEVICE_NAME)) == NULL) {
			fprintf(stderr, "failed to get scsi device name\n");
			return 1;
		}
		run_scsi_diag(dev_num);
		free(dev_num);
		return 0;
	}
	if (strncmp(BUS_NAME, "pci", 3) == 0) {
		if (DIAG_OPTIONS & DIAG_ALL)
			return (run_net_diag_all());
		if ((dev_num = get_net_dev(dev_num)) == NULL)
			return 1;
		run_net_diag(dev_num);
		free(dev_num);
		return 0;
	}
	return 1;
}


int run_diag_class()
{
	char *dev_num = NULL;
	char *path = NULL;
	if (strncmp(CLASS_NAME, "net", 3) == 0) {
		if (DIAG_OPTIONS & DIAG_ALL) {
			return (run_net_diag_all());
		}
		if ((dev_num = get_net_dev(dev_num)) == NULL) 
			return 1;
		run_net_diag(dev_num);
		free(dev_num);
		return 0;
	} else {
		if ((strncmp(CLASS_NAME, "scsi_device", 11) == 0) || 
				(strncmp(CLASS_NAME, "scsi_host", 9) == 0)) {
			if (DIAG_OPTIONS & DIAG_ALL) {
				if (DIAG_OPTIONS & DIAG_LOGS) {
	fprintf(stdout, "\n-------------------------------------\n");
	fprintf(stdout, "Getting error logs for SCSI\n");
	fprintf(stdout, "-------------------------------------\n");
					DIAG_OPTIONS &= ~DIAG_LOGS;
					path = (char *)show_logs("scsi");
					if (path != NULL) {
						fflush(stdout);
						system(path);
					}
				}
				return (run_scsi_diag_all());
			}
		}
		if (strncmp(DEVICE_NAME, "host", 4) == 0) {
			if ((dev_num = get_def_scsi_host_dev(DEVICE_NAME))
					== NULL) 
				return 1;
		} else {
			if ((dev_num = get_scsi_quad((char *)DEVICE_NAME)) 
					== NULL) {
				fprintf(stderr, 
					"failed to get scsi device name\n");
				return 1;
			}
		}
		run_scsi_diag(dev_num);
		free(dev_num);
		return 0;
	}
	return 1;
}


/**
 * This function will parse the dev-class and dev-num and call
 * the appropriate function to show the device information.
 *
 * for sysdiag -i net
 * this function will look for the firt device in the net dir or the
 * /sysfs/class. and run ethtool -i on that.
 *
 * For ethernet device, in the /sys/class/net class, we look for the
 * ethX device and do the diag on it.
 *
 * For scsi device, in the /sys/class/scsi_host  dir, we go thro the
 * hostX dir and then in the device dir look at the 0:0:0:0 and then
 * look the link info for block and then see device type, map it to
 * the sgX device and then do the diagnostics on that.
 *
 * And if the scsi_host class is passed then it will call the sg_inq
 * function to show the information for that scsi_device.
 *
 *
 * Format for dev-info is
 * #sysdiag -i [dev-class] [dev-num]
 *
 * dev-class	: scsi_host, net.
 * dev-num	: eth0, sg0, 0:0:0:0.
 *
 * format for run_diag is::
 * sysdiag -i -b pci -D 0000:00:00.0
 * sysdiag -i -b pci -A
 * sysdiag -i -c net [-D eth0] or -A
 * sysdiag -i -c scsi_host [-D 0:0:0:0 or sg0 or sda] or -A
 *
 * sysdiag -a eth0
 * sysdiag -i eth0
 * sysdiag net eth0
 * 
 *
 * This function will parse the dev-class and dev-num and call
 * the appropriate function to run diagnostics on the appropriate
 * device.
 */
int run_diag()
{
	char *path = NULL;
	if (DIAG_OPTIONS & DIAG_CLASS) {
		run_diag_class();
		return 0;
	}else if (DIAG_OPTIONS & DIAG_BUS) {
		run_diag_bus();
		return 0;
	}
	if (DIAG_OPTIONS & DIAG_ALL) {
		if (DIAG_OPTIONS & DIAG_LOGS) {
			fprintf(stdout, "\n-------------------------------------\n");
			fprintf(stdout, "Getting error logs for SCSI\n");
			fprintf(stdout, "-------------------------------------\n");
			DIAG_OPTIONS &= ~DIAG_LOGS;
			path = (char *)show_logs("scsi");
			if (path != NULL) {
				fflush(stdout);				
				system(path);
			}
			run_scsi_diag_all();
			DIAG_OPTIONS |= DIAG_LOGS;
		} else 
			run_scsi_diag_all();
		run_net_diag_all();
		return 0;
	}
	return 0;
}


int search_links(char *path, char *device)
{
	struct sysfs_directory *dir = NULL;
	struct sysfs_link *link = NULL;
	char name[SYSFS_NAME_LEN];
	int ret;
	char *path1;
	struct stat statbuf;

	path1 = (char *)calloc(SYSFS_PATH_MAX, sizeof(char));
	memset(name, 0, SYSFS_NAME_LEN);

	if (path == NULL) {
		path = (char *)calloc(SYSFS_PATH_MAX, sizeof(char));
		strncpy(path, SYSFS_MOUNT_PATH, strlen(SYSFS_MOUNT_PATH));
	}

	strncpy(path1, path, strlen(path));
	strncat(path1, "/devices", strlen("/devices"));
	if (stat(path1, &statbuf) != 0) { 
		free (path1);
		return 0;
	}
	if ((dir = sysfs_open_directory(path1)) == NULL) {
		fprintf(stdout, "error opening dir %s\n", path1);
		free (path1);
		return 0;
	}
	free(path1);

	if (sysfs_read_directory(dir) != 0) {
		fprintf(stdout, "error reading dir %s\n", path1);
		return 0;
	}
	ret = 0;
	for (link = dir->links; link != NULL; link = link->next) {
		if (strcmp(link->name, device) == 0) {
			if ((sysfs_get_name_from_path(path, name,
						SYSFS_NAME_LEN)) != 0) {
				sysfs_close_directory(dir);
				return 1;
			}
			sysfs_close_directory(dir);
			return -1;
		}
	}
        sysfs_close_directory(dir);
        return 0;
}

/**
 * to find the path to the device passed, we need to open the sysfs dir and
 * look for the device in there,
 *
 * we do opendir on sys/ and then readdir,
 * 	if the dir != device, opendir(dir) then readdir again,
 * 	if element != device, then keep looking for it.
 *
 *
 */

int search_subdirs(char *path,char *device)
{
	struct sysfs_directory *dir = NULL, *current = NULL, *sdir = NULL;
	struct sysfs_link *lnk = NULL ;
	char name[SYSFS_NAME_LEN], type[SYSFS_NAME_LEN];
	int ret;
	struct stat statbuf;

	memset(name, 0, SYSFS_NAME_LEN);
	memset(type, 0, SYSFS_NAME_LEN);

	if (path == NULL) {
		path = (char *)calloc(SYSFS_PATH_MAX, sizeof(char));
		strncpy(path, SYSFS_MOUNT_PATH, strlen(SYSFS_MOUNT_PATH));
	}
	if (stat(path,&statbuf) != 0) 
		return 0;
	if ((dir = sysfs_open_directory(path)) == NULL) {
		fprintf(stdout,"error opening dir %s\n",path);
		return 0;
	}
	if (sysfs_read_directory(dir) != 0) {
		fprintf(stdout,"error reading dir %s\n",path);
		return 0;
	}
	current = dir->subdirs;
	ret = 0;
	while (current != NULL) {
		if ((sysfs_get_name_from_path(current->path, name,
						SYSFS_NAME_LEN)) != 0) {
			sysfs_close_directory(dir);
			return 0;
		}
		if (strcmp(name,device) == 0) {
			if (strstr(current->path, "drivers") != NULL) {
				if ((sdir = sysfs_open_directory(current->path))
								== NULL) {
				fprintf(stdout,"error opening dir %s\n",path);
					return 0;
				}
				if (sysfs_read_directory(sdir) != 0) {
					fprintf(stdout,"error reading dir %s\n"
									,path);
					return 0;
				}
				lnk = sdir->links;
			}
			ret = -1;
			break;
		}
		if ((ret = search_subdirs(current->path, device)) != 0) { 
			if (strstr(current->path, "drivers") == NULL &&
								(ret != 2)) {
				memset(path, 0, SYSFS_NAME_LEN);
				if ((sysfs_get_name_from_path(current->path, 
						path, SYSFS_NAME_LEN)) != 0) {
					sysfs_close_directory(dir);
					return 0;
				}
				ret = 1;
			}
			else { 
				if (ret != 2) {
					memset(name, 0, SYSFS_NAME_LEN);
					if ((sysfs_get_name_from_path(
							path, 
							path, 
							SYSFS_NAME_LEN)) != 0) {
						sysfs_close_directory(dir);
						return 0;
					}
					memset(BUS_NAME, 0, SYSFS_NAME_LEN);
					strncpy(BUS_NAME, path, strlen(path));
				}
				ret = 2;
			}
			sysfs_close_directory(dir);
			return ret;
		}
		current = current->next;
	}
	for (current = dir->subdirs; current != NULL; current = current->next)
	{
		if (search_links(current->path, device) == -1) {
			memset(path, 0, SYSFS_NAME_LEN);
			if ((sysfs_get_name_from_path(current->path, path,
						SYSFS_NAME_LEN)) != 0) {
				sysfs_close_directory(dir);
				return 0;
			}
			ret = 1;
			break;
		}
	}
	sysfs_close_directory(dir);
        return ret;
}

/**
 * here we check if the device passed is a root device.
 *
 */

int find_root(char *device)
{
	char path[SYSFS_NAME_LEN];
	char name[SYSFS_NAME_LEN];
	struct sysfs_directory *dir = NULL, *curr = NULL;

	memset(path, 0, SYSFS_NAME_LEN);
	strncpy(path, SYSFS_MOUNT_PATH, strlen(SYSFS_MOUNT_PATH));
	strncat(path, "/devices", 8);

	dir = sysfs_open_directory(path);
	if (dir == NULL) {
		fprintf(stderr, "Error opening sysfs directory at %s\n", path);
		return 1;
	}
	if ((sysfs_read_directory(dir)) != 0) {
		fprintf(stderr, "Error reading sysfs directory at %s\n", path);
		return 1;
	}
	curr = dir->subdirs;
	while (curr != NULL) {
		if ((sysfs_get_name_from_path(curr->path, name,
						SYSFS_NAME_LEN)) != 0) {
			sysfs_close_directory(dir);
			return 1;
		}
		if (strcmp(name, device) == 0) {
			strncpy(TOPO_NAME, device, strlen(device));
			DIAG_OPTIONS |= DIAG_TOPO;
			sysfs_close_directory(dir);
			return 0;
		}
		curr = curr->next;
	}
	sysfs_close_directory(dir);
	return 1;

}

/**
 * to find the device, first look for BUS or CLASS or DRIVER_NAME
 * if not found,
 * Then opendir( sys/bus/ ), readdir();
 * for each subdir -> opendir( subdir+devices) then look for devices there
 * in the links.
 *
 *
 */

int find_device(int argc, char *argv[])
{	
	int i=0, ret = 0;
	char class_path[SYSFS_PATH_MAX];
	char bus_path[SYSFS_PATH_MAX];

	i = argc;
	if (argc == 0) return 0;
	
	for (i = argc - 1 ; i >=0 ; --i) {
		
		memset(bus_path, 0, SYSFS_NAME_LEN);
		memset(class_path, 0, SYSFS_NAME_LEN);
	
		strncpy(bus_path, SYSFS_MOUNT_PATH, strlen(SYSFS_MOUNT_PATH));
		strncat(bus_path, SYSFS_BUS_DIR, strlen(SYSFS_BUS_DIR));
		strncpy(class_path, SYSFS_MOUNT_PATH, strlen(SYSFS_MOUNT_PATH));
		strncat(class_path, SYSFS_CLASS_DIR, strlen(SYSFS_CLASS_DIR));
		switch (search_subdirs(bus_path, argv[i])) {
		case 1:
			strncpy(DEVICE_NAME, argv[i], strlen(argv[i]));
			strncpy(BUS_NAME, bus_path, strlen(bus_path));
			DIAG_OPTIONS |= DIAG_BUS | DIAG_DEVICE;
			ret = 0;
			break;
		case 2:
			strncpy(DRIVER_NAME, argv[i], strlen(argv[i]));
			DIAG_OPTIONS |= DIAG_BUS | DIAG_DRIVER ;
			ret = 0;
			break;
		case -1:
			strncpy(BUS_NAME, argv[i], strlen(argv[i]));
			DIAG_OPTIONS |= DIAG_BUS;
			ret = 0;
			break;
		default:
			switch(search_subdirs(class_path, argv[i])) {
			case 1:
				strncpy(DEVICE_NAME, argv[i], strlen(argv[i]));
				strncpy(CLASS_NAME, class_path, 
							strlen(class_path));
				DIAG_OPTIONS |= DIAG_CLASS | DIAG_DEVICE;
				ret =0;
				break;
			case -1:
				strncpy(CLASS_NAME, argv[i], strlen(argv[i]));
				DIAG_OPTIONS |= DIAG_CLASS;
				ret = 0;
				break;
			default:
				if (find_root(argv[i]) != 0) {
					get_scsi_quad(argv[i]);
					if (DEVICE_NAME[0] == 0) {
						fprintf(stderr, 
						"invalid argument: %s\n", 
						argv[i]);
						return -1;
					}
				}
				break;
			}
			break;
		}
	}
	return ret;
}

/**
 * main function.
 *
 * sysdiag <options> [bus|class] [device]
 *
 * sysdiag -a eth0
 * sysdiag -i eth0
 * sysdiag net eth0
 *
 */
int main(int argc, char *argv[])
{
	int opt;
	extern char *optarg;
	extern int optind, opterr, optopt;
	DIAG_LEVEL = 0;
	DIAG_OPTIONS = 0;

	memset(BUS_NAME, 0, SYSFS_NAME_LEN);
	memset(TOPO_NAME, 0, SYSFS_NAME_LEN);
	memset(CLASS_NAME, 0, SYSFS_NAME_LEN);
	memset(DEVICE_NAME, 0, SYSFS_NAME_LEN);
	memset(DRIVER_NAME, 0, SYSFS_NAME_LEN);

	while ((opt = getopt(argc,argv,CMD_OPTIONS)) != -1) {
		switch (opt) {
			case 'b':
				if (optarg == 0) {
					usage();
					exit(0);
				}
				strncpy(BUS_NAME, optarg, strlen(optarg));
				DIAG_OPTIONS |= DIAG_BUS;
				break;
			case 'c':
				if (optarg == 0) {
					usage();
					exit(0);
				}
				strncpy(CLASS_NAME, optarg, strlen(optarg));
				DIAG_OPTIONS |= DIAG_CLASS;
				break;
			case 'i':
				DIAG_OPTIONS |= DIAG_INQ;
				break;
			case 'd':
				DIAG_OPTIONS |= DIAG_DIAG_INFO;
				
				break;
			case 'l':
				DIAG_OPTIONS |= DIAG_LOGS;
				break;
			case 'p':
				DIAG_OPTIONS |= DIAG_PAUSE;
				break;
			case 's':
				DIAG_OPTIONS |= DIAG_STAT;
				break;
			case 't':
				DIAG_OPTIONS |= DIAG_TEST;
				break;
			case 'r':
				if (optarg == 0) {
					usage();
					exit(0);
				}
				strncpy(TOPO_NAME, optarg, strlen(optarg));
				DIAG_OPTIONS |= DIAG_TOPO;
				break;
			case 'v':
				DIAG_OPTIONS |= DIAG_SHOW_VALUE 
					     | DIAG_LIST;
				DIAG_LEVEL++;
				break;
			case 'x':
				DIAG_OPTIONS |= DIAG_ALL;
				break;
			case 'h':
				usage();
				exit(0);
			default:
				usage();
				exit(0);
		}
	}
	argc -= optind;
	argv += optind;

	if (DIAG_OPTIONS & DIAG_DIAG_INFO) {
		show_diag_info(argv[0]);
	}
	if (setup_sysfs_paths() != 0) {
		fprintf(stderr, "sysfs not found\n");
		exit(0);
	}
	if (find_device(argc,argv) == -1) 
			exit(1);
		
	if ((DIAG_OPTIONS & DIAG_INQ)  ||
	    (DIAG_OPTIONS & DIAG_TEST) ||
	    (DIAG_OPTIONS & DIAG_LOGS) ||
	    (DIAG_OPTIONS & DIAG_STAT))
		run_diag();
	else
		if (!(DIAG_OPTIONS & DIAG_DIAG_INFO))
			DIAG_OPTIONS |= DIAG_LIST | DIAG_DEVICE | DIAG_DRIVER;

	if (DIAG_OPTIONS & DIAG_LIST) {
		list_main();
	}
	if (DIAG_OPTIONS & DIAG_TOPO) {  
		show_sysfs_root(TOPO_NAME);
	}
	return 0;

} /* end main */
